<?php

/*
+---------------------------------------------------------------------------+
| Max Media Manager v0.3                                                    |
| =================                                                         |
|                                                                           |
| Copyright (c) 2003-2005 m3 Media Services Limited                         |
| For contact details, see: http://www.m3.net/                              |
|                                                                           |
| This program is free software; you can redistribute it and/or modify      |
| it under the terms of the GNU General Public License as published by      |
| the Free Software Foundation; either version 2 of the License, or         |
| (at your option) any later version.                                       |
|                                                                           |
| This program is distributed in the hope that it will be useful,           |
| but WITHOUT ANY WARRANTY; without even the implied warranty of            |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
| GNU General Public License for more details.                              |
|                                                                           |
| You should have received a copy of the GNU General Public License         |
| along with this program; if not, write to the Free Software               |
| Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA |
+---------------------------------------------------------------------------+
$Id: FilterCompensation.php 3819 2005-10-27 20:17:00Z andrew@m3.net $
*/

require_once MAX_PATH . '/lib/max/Maintenance/Priority/AdServer/Task.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Entities.php';

/**
 * The minimum difference in the fraction of a zone's actual
 * impressions that an ad is given before filter compensation
 * will be applied again -- that is, if a filter compensation
 * step doesn't change the ad's fraction of impressions by
 * more than this amount, then don't try to compensate any
 * further.
 */
define('ZONE_TRAFFIC_FRACTION_MINIMUM_DELTA', 0.0001);

/**
 * A class to carry out the task of compensating for the effect of filters
 * on advertisement/banner priorities.
 *
 * @package    MaxMaintenance
 * @subpackage Priority
 * @author     Andrew Hill <andrew@m3.net>
 */
class FilterCompensation extends MAX_Maintenance_Priority_AdServer_Task
{
    
    /**
     * The main method of the class, that is run by the controlling 
     * task runner class.
     */
    function run()
    {
        $aPriorities = array();
        // Get all zone/ad information
        $aZones = &$this->_buildClasses();
        // For every zone with ads linked to it...
        if (count($aZones) > 0) {
            foreach ($aZones as $oZone) {
                if (!$oZone->containsFilteredAds()) {
                    // There are no filtered ads in the zone, so calculate
                    // the priorities based on the the unmodified required
                    // impression values
                    $aPriorities[$oZone->id] = $this->initialPriorities($oZone);
                } else {
                    // Calculate the priorities based on the required impression
                    // values AND the past information about previous priorities
                    $aPriorities[$oZone->id] = $this->learnedPriorities($oZone);
                }
            }
            // Store the calculated priorities
            $this->oDal->updatePriorities($aPriorities);
        }
    }
    
    /**
     * A private method to construct the classes used in performing
     * filter compensation.
     *
     * @access private
     * @return mixed An array of Zone objects, each populated with the
     *               relevant Advert objects and data.
     */
    function &_buildClasses()
    {
        // Obtain the forecast impression inventory for each zone for the current OI
        $aZoneImpInvs = &$this->oDal->getAllZonesImpInv();
        // Create an array of all of the zones, indexed by zone ID
        $aZones = array();
        if (is_array($aZoneImpInvs) && (count($aZoneImpInvs) > 0)) {
            foreach ($aZoneImpInvs as $aZoneImpInv) {
                // Create and store the Zone object
                $aZones[$aZoneImpInv['zone_id']] = &new Zone(array('zoneid' => $aZoneImpInv['zone_id']));
                // Record the zone's forecast impression inventory
                $aZones[$aZoneImpInv['zone_id']]->availableImpressions = $aZoneImpInv['forecast_impressions'];
                // Record the zone's previous operation interval actual impressions
                $aZones[$aZoneImpInv['zone_id']]->pastActualImpressions = $aZoneImpInv['actual_impressions'];
            }
        }
        // Obtain the ad/zone combinations where advertisements have had
        // impressions allocated to zones
        $aZoneImpAllocs = &$this->oDal->getAllZonesWithAllocInv();
        // Obtain the details of advertisements that are filtered
        $aFilteredAds = &$this->_getFilteredAds();
        // Add Advert objects to the zones
        if (is_array($aZoneImpAllocs) && (count($aZoneImpAllocs) > 0)) {
            foreach ($aZoneImpAllocs as $aZoneImpAlloc) {
                // Create an Advert object
                $oAdvert = &new Advert(array('ad_id' => $aZoneImpAlloc['ad_id']));
                // Assign the required impressions for this ad/zone
                $oAdvert->requiredImpressions = $aZoneImpAlloc['impressions'];
                // If the ad is filtered, set the Advert's filtered flag
                if ($aFilteredAds[$oAdvert->id] === true) {
                    $oAdvert->filtered = true;
                }
                // Add the Advert object to the appropriate Zone object
                $aZones[$aZoneImpAlloc['zone_id']]->addAdvert($oAdvert);
            }
        }
        // Get the details of the previous requested/delivered ad impressions and
        // calculated priorities
        $aPastDetails = &$this->oDal->getLastOperationIntervalAdDeliveryInfo();
        if (is_array($aPastDetails) && (count($aPastDetails) > 0)) {
            foreach ($aPastDetails as $aPastDetail) {
                // Only insert the past details for ads that are still linked to the same zone
                if (isset($aZones[$aPastDetail['zone_id']]->aAdverts[$aPastDetail['ad_id']])) {
                    $aZones[$aPastDetail['zone_id']]->aAdverts[$aPastDetail['ad_id']]->pastRequiredImpressions
                        = $aPastDetail['impressions_requested'];
                    $aZones[$aPastDetail['zone_id']]->aAdverts[$aPastDetail['ad_id']]->pastActualImpressions
                        = $aPastDetail['impressions'];
                    $aZones[$aPastDetail['zone_id']]->aAdverts[$aPastDetail['ad_id']]->pastAdZonePriorityFactor
                        = $aPastDetail['priority_factor'];
                    $aZones[$aPastDetail['zone_id']]->aAdverts[$aPastDetail['ad_id']]->pastZoneTrafficFraction
                        = $aPastDetail['past_zone_traffic_fraction'];
                }
            }
        }
        return $aZones;
    }
    
    /**
     * A private method to return an array of all active, filtered advertisements.
     *
     * @access private
     * @return array An array, indexed by advertisement ID, where the value is true
     *               if the advertisement is a filtered ad.
     */
    function &_getFilteredAds()
    {
        // Get all ads with delivery limitations
        $aLimitedAds = $this->oDal->getAllDeliveryLimitedAds();
        // Create the array 
        $aFilteredAds = array();
        foreach ($aLimitedAds as $aLimitedAd) {
            // Is this a filtered ad?
            if ($aLimitedAd['block'] > 0) {
                $aFilteredAds[$aLimitedAd['bannerid']] = true;
                continue;
            }
            if ($aLimitedAd['capping'] > 0) {
                $aFilteredAds[$aLimitedAd['bannerid']] = true;
                continue;
            }
            if ($aLimitedAd['session_capping'] > 0) {
                $aFilteredAds[$aLimitedAd['bannerid']] = true;
                continue;
            }
            if ($this->_isFilteringLimitation($aLimitedAd['compiledlimitation'])) {
                $aFilteredAds[$aLimitedAd['bannerid']] = true;
            }
        }
        return $aFilteredAds;
    }
    
    /**
     * A private method for testing advertisement compiled limitations to determine if
     * any of the limitations are blocking - that is...
     *
     * @access private
     * @param string $limitation The compiled limitation.
     * @return boolean True if the limitation is a filtering limitation, false
     *                 otherwise.
     */
    function _isFilteringLimitation($limitation)
    {
        // Is the limitation true?
        if ($limitation == 'true') {
            return false;
        }
        // Does the limitation have phpAds_aclCheck()?
        if (preg_match("/phpAds_aclCheck\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckChannel()?
        if (preg_match("/phpAds_aclCheckChannel\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckUseragent()?
        if (preg_match("/phpAds_aclCheckUseragent\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckClientip()?
        if (preg_match("/phpAds_aclCheckClientip\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckDomain()?
        if (preg_match("/phpAds_aclCheckDomain\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckLanguage()?
        if (preg_match("/phpAds_aclCheckLanguage\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckSource()?
        if (preg_match("/phpAds_aclCheckSource\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckCountryCode()?
        if (preg_match("/phpAds_aclCheckCountryCode\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckCountryName()?
        if (preg_match("/phpAds_aclCheckCountryName\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckRegion()?
        if (preg_match("/phpAds_aclCheckRegion\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckCity()?
        if (preg_match("/phpAds_aclCheckCity\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckPostalCode()?
        if (preg_match("/phpAds_aclCheckPostalCode\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckPostalLatitude()?
        if (preg_match("/phpAds_aclCheckPostalLatitude\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckPostalLongitude()?
        if (preg_match("/phpAds_aclCheckPostalLongitude\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckDmaCode()?
        if (preg_match("/phpAds_aclCheckDmaCode\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckAreaCode()?
        if (preg_match("/phpAds_aclCheckAreaCode\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckAreaOrganisation()?
        if (preg_match("/phpAds_aclCheckAreaOrganisation\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckAreaNetspeed()?
        if (preg_match("/phpAds_aclCheckAreaNetspeed\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckContinent()?
        if (preg_match("/phpAds_aclCheckContinent\([^)]+\)/", $limitation)) {
            return true;
        }
        // Does the limitation have phpAds_aclCheckReferer()?
        if (preg_match("/phpAds_aclCheckReferer\([^)]+\)/", $limitation)) {
            return true;
        }
        return false;
    }
    
    /**
     * A method for scaling the advertisement and blank priorities in a zone
     * so that the sum of the priorities is unity.
     *
     * Note that the input values are normally simply the required number
     * of impressions for each advertisement, and the number of remaining
     * impressions in the zone.
     *
     * @param mixed $aParams A hash with with following indexes:
     *                 - "ads" An array (indexed by ad id) of hashes,
     *                         each hash containing the following
     *                         indexes/data:
     *                          - "ad_id"                 The ad ID.
     *                          - "zone_id"               The zone ID.
     *                          - "priority"              The ad/zone priority value.
     *                          - "requested_impressions" The number of impressions the priority
     *                                                    should result in.
     *                 - "blank" The blank priority for the zone.
     */
    function scalePriorities(&$aParams, $adjust = false)
    {
        // Find sum of values
        $total = 0;
        foreach ($aParams['ads'] as $ad) {
            $total += $ad['priority'];
        }
        $total += $aParams['blank'];
        // Ensure the total is > 0
        if ($total <= 0) {
            // Cannot scale, return original values
            return;
        }
        // Scale non-blank priorities
        foreach ($aParams['ads'] as $key => $ad) {
            $aParams['ads'][$key]['priority'] /= $total;
        }
        // Scale blank priority
        $aParams['blank'] /= $total;
    }
    
    /**
     * A method for creating an array of initial priority values.
     *
     * @param Zone $oZone A Zone object, containing the relevant zone
     *                    data and associated Adverts.
     * @return mixed A hash with with following indexes:
     *                 - "ads" An array (indexed by ad id) of hashes, each
     *                   hash containing the following indexes/data:
     *                          - "ad_id"                 The ad ID.
     *                          - "zone_id"               The zone ID.
     *                          - "priority"              The ad/zone priority value.
     *                          - "requested_impressions" The number of impressions the priority
     *                                                    should result in.
     *                 - "blank" The blank priority for the zone.
     */
    function initialPriorities($oZone)
    {
        $result = array('ads' => array());
        // Test available zone impressions > 0
        if ($oZone->availableImpressions <= 0) {
            // There are no zone impressions, so there cannot be any
            // priorities, so assign each ad a priority of zero, and
            // assign 1 as the initial blank priority
            foreach ($oZone->aAdverts as $oAdvert) {
                $result['ads'][$oAdvert->id] = array(
                    'ad_id'                 => $oAdvert->id,
                    'zone_id'               => $oZone->id,
                    'priority'              => 0,
                    'impressions_requested' => $oAdvert->requiredImpressions
                );
            }
            $result['blank'] = 1;
            return $result;
        }
        // For each advertisement in the zone, calculate the initial
        // priority as the number of required impressions for each
        // advertisement divided by the total number of available
        // impressions in the zone. Increment a counter of the number
        // of allocated impressions as this is done.
        $usedImpressions = 0;
        foreach ($oZone->aAdverts as $oAdvert) {            
            $result['ads'][$oAdvert->id] = array(
                'ad_id'                 => $oAdvert->id,
                'zone_id'               => $oZone->id,
                'priority'              => $oAdvert->requiredImpressions
                                           / $oZone->availableImpressions,
                'impressions_requested' => $oAdvert->requiredImpressions
            );
            $usedImpressions += $oAdvert->requiredImpressions;
        }
        // Were there more impressions allocated than exist?
        if ($usedImpressions >= $oZone->availableImpressions) {
            // Initial blank priority of zero
            $result['blank'] = 0;
        } else {
            // Initial blank priority of remaining impressions divided
            // by the total number of available impressions.
            $result['blank'] =
                ($oZone->availableImpressions - $usedImpressions)
                / $oZone->availableImpressions;
        }
        // Scale the priority values
        $this->scalePriorities($result);
        return $result;
    }
    
    /**
     * A method for creating an array of learned priority values.
     *
     * @param Zone $oZone A Zone object, containing the relevant zone
     *                    data and associated Adverts.
     * @return mixed A hash with with following indexes:
     *                 - "ads" An array (indexed by ad id) of hashes, each
     *                   hash containing the following indexes/data:
     *                          - "ad_id"                       The ad ID.
     *                          - "zone_id"                     The zone ID.
     *                          - "priority"                    The ad/zone priority value.
     *                          - "requested_impressions"       The number of impressions the priority
     *                                                          should result in.
     *                          - "priority_factor"             The adjustment factor applied to the
     *                                                          ad's priority to compensate for filters.
     *                          - "past_zone_traffic_fraction"  The fraction of the zone's impressions
     *                                                          given to the ad in the previous operation
     *                                                          interval.
     *                 - "blank" The blank priority for the zone.
     */
    function learnedPriorities($oZone)
    {
        // Calculate the priority values the normal way
        $result = $this->initialPriorities($oZone);
        // Adjust each ad's priority value based on past data
        foreach ($oZone->aAdverts as $oAdvert) {
            // Calculate the ad's priority adjustment factor
            list($factor, $fraction) =
                $this->_getPriorityAdjustment($oAdvert,
                                              $oZone->pastActualImpressions);
            // Adjust the ad's priority
            $result['ads'][$oAdvert->id]['priority'] *= $factor;
            // Store the ad's priority adjustment factor
            $result['ads'][$oAdvert->id]['priority_factor'] = $factor;
            // Store the fraction of the zone's impressions used by the
            // ad in the previous operation interval
            $result['ads'][$oAdvert->id]['past_zone_traffic_fraction'] = $fraction;
        }
        // Re-scale the priority values
        $this->scalePriorities($result);
        return $result;
    }
    
    /**
     * A private function for calculating how much an advert's priority
     * value needs to be adjusted.
     *
     * @access private
     * @param Advert $oAdvert An Advert object.
     * @param integer $zoneImpressions The total number of (actual) impressions in the 
     *                                 zone (for which the ad's priority is being
     *                                 adjusted) in the previous operation interval.
     * @return array An array containing:
     *                  - The ad's priority adjustment factor
     *                  - The fraction of the zone's impressions given to the ad
     *                    in the previous operation interval.
     */
    function _getPriorityAdjustment($oAdvert, $zoneImpressions)
    {
        if (!is_null($oAdvert->pastRequiredImpressions) &&
                !is_null($oAdvert->pastActualImpressions) &&
                ($oAdvert->pastRequiredImpressions != 0) &&
                ($oAdvert->pastActualImpressions != 0)) {
            // Calculate the fraction of the zone traffic that was seen
            // by the advertisement in the past interval
            if ($zoneImpressions > 0) {
                $fraction = $oAdvert->pastActualImpressions / $zoneImpressions;
            } else {
                unset($fraction);
            }
            // Is this fraction much different to the fraction before that,
            // given that a past fraction exists, and the ad has had filter
            // compensation priority adjustment factor applied before?
            if (!is_null($oAdvert->pastZoneTrafficFraction) &&
                !is_null($oAdvert->pastAdZonePriorityFactor) &&
                !is_null($fraction)) {
                $difference = abs($fraction - $oAdvert->pastZoneTrafficFraction);
                if ($difference < ZONE_TRAFFIC_FRACTION_MINIMUM_DELTA) {
                    // Don't bother with additional filter compensation, assuming
                    // we want to adjust the priority in the same direction as
                    // before
                    $factor = $this->_calculateFactor($oAdvert->pastAdZonePriorityFactor,
                                                      $oAdvert->pastRequiredImpressions,
                                                      $oAdvert->pastActualImpressions,
                                                      false);
                    return array($factor, $fraction);
                }
            }
            // Any previous filter compensation had an affect, do we want to
            // try even more priority adjustment?
            if (!is_null($oAdvert->pastAdZonePriorityFactor)) {
                $factor = $this->_calculateFactor($oAdvert->pastAdZonePriorityFactor,
                                                  $oAdvert->pastRequiredImpressions,
                                                  $oAdvert->pastActualImpressions,
                                                  true);
                return array($factor, $fraction);
            }
            // We can't adjust the priority any more, or no previous adjustment,
            // so perform basic adjustment based on the past operation interval's
            // performance
            $factor = $oAdvert->pastRequiredImpressions / $oAdvert->pastActualImpressions;
            return array($factor, $fraction);
        } else {
            // No past information, so use a factor of 1
            return array(1, null);
        }
    }
    
    /**
     * A private function for calculating the prioritisation adjustment factor
     * to use.
     *
     * @param double $oldFactor  The previous priority adjustment factor
     * @param integer $required  The number of impressions required in the past interval.
     * @param integer $delivered The number of impressions delivered in the past interval.
     * @param boolean $useNew    True if new new factor should be used when valid, false
     *                           to use the previous factor when valid. A new base factor
     *                           is used regardless in the case that either the new factor
     *                           or the previous factor are not valid.
     * @return double The priority adjustment factor to use.
     */
    function _calculateFactor($oldFactor, $required, $delivered, $useNew)
    {
        // If no impressions were delivered last interval, can't adjust priority
        if ($delivered == 0) {
            return 1;
        }
        // Calculate the base priority adjustment factor, on the basis of the
        // past interval's performance
        $baseFactor = $required / $delivered;
        // If the base factor is unity, then the factor from the last interval
        // was exactly, right, so don't bother calculating the factor again
        if ($baseFactor == 1) {
            return $oldFactor;
        }
        // Calculate the full priority adjustment factor, on the basis of the
        // product of the past factor and the new base factor
        $fullFactor = $oldFactor * $baseFactor;
        // Is the full factor a real number?
        if (is_infinite($fullFactor) || is_nan($fullFactor)) {
            // Unable to calculate the full factor, so just use the factor
            // from the previous interval as the new factor
            $newFactor = $oldFactor;
        } else {
            // Use the full factor as the new factor
            $newFactor = $fullFactor;
        }
        if ($baseFactor > 1) {
            // The base factor is trying to increase priority
            if ($newFactor > 1) {
                // New factor is also trying to increase priority
                if ($useNew) {
                    // Use the new factor
                    return $newFactor;
                } else {
                    // Use the old factor
                    return $oldFactor;
                }
            } else {
                // The new factor is actually trying to lower the priority
                if ($useNew) {
                    // Use the new factor
                    return $newFactor;
                } else {
                    // Use the (new) base factor
                    return $baseFactor;
                }
            }
        }
        if ($baseFactor < 1) {
            // The base factor is trying to decrease priority
            if ($newFactor < 1) {
                // New factor is also trying to decrease priority
                if ($useNew) {
                    // Use the new factor
                    return $newFactor;
                } else {
                    // Use the old factor
                    return $oldFactor;
                }
            } else {
                // The new factor is actually trying to raise the priority
                if ($useNew) {
                    // Use the new factor
                    return $newFactor;
                } else {
                    // Use the (new) base factor
                    return $baseFactor;
                }
            }
        }
    }
    
}
                                      
?>
